#include <dlfcn.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <time.h>
#include "f1.h"

#define FUNCTION_TO_LOAD "f2"

#ifdef __cplusplus
extern "C"{
#endif

extern char *getString();


char sg[300];
void *library;
char *lib_path;
bool isRun;
int (*function)(void);
pthread_t *th;

void sig_handler(int signum){
    switch(signum){
    case SIGTERM:
        break;
    }
    fprintf(stderr, "Number of signal: %d\n", signum);
}

void changeString(char *ns){
    strcpy(sg, ns);
}

static void __attribute__((constructor)) init(){
    fprintf(stderr, "Constructor started\n");
    isRun = false;
    library = NULL;
    th = NULL;
    strcpy(sg, "f1 string");
}

static void __attribute__((destructor)) finit(){
    fprintf(stderr, "Destructor started\n");
    isRun = false;
    sg[0] = '\0';
    if (library != NULL){
        unloadLibrary();
    }
    if (lib_path != NULL){
        free(lib_path);
        lib_path = NULL;
    }
    if (th != NULL){
        free(th);
        th = NULL;
    }
}

static void * run_th(void *args){
    isRun = true;
    signal(SIGTERM, sig_handler);
    function();
    isRun = false;
    return NULL;
}


bool loadLibrary(const char * const s){
    lib_path = strdup(s);
    library = dlopen(lib_path, RTLD_LAZY);
    if (!library){
        fprintf(stderr, "dlopen() error: %s\n", dlerror());
        return false;
    }
    function = (int (*)(void))dlsym(library, FUNCTION_TO_LOAD);
    if (!function){
        fprintf(stderr, "dlsym() error: %s\n", dlerror());
        unloadLibrary();
        return false;
    }
    return true;
}

bool unloadLibrary(){
    if (library == NULL){
        return false;
    }
    int rt = dlclose(library);
    if (rt != 0){
        fprintf(stderr, "dlclose() error: %s\n", dlerror());
    }
    library = NULL;
    if (lib_path != NULL){
        free(lib_path);
        lib_path = NULL;
    }
    if (th != NULL){
        free(th);
        th = NULL;
    }
    return (rt == 0);
}

bool stopExecution(){
    if (!isRun){
        return true;
    }
    if (th == NULL){
        isRun = false;
        return true;
    }
    int state;
    state = pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
    if (state != 0){
        fprintf(stderr, "pthread_setcancelstate\n");
        return false;
    }
    state = pthread_kill(*th, SIGTERM);
    if (state != 0){
        fprintf(stderr, "pthread_cancel\n");
        return false;
    }
    void *res;
    state = pthread_join(*th, &res);
    if (state != 0){
        fprintf(stderr, "pthread_join\n");
    }
    if (res == PTHREAD_CANCELED){
        fprintf(stderr, "thread was canceled\n");
    }else{
        fprintf(stderr, "thread wasn't canceled\n");
    }
    free(th);
    th = NULL;
    isRun = false;
    return true;
}

bool startExecution(long long ms){
    if (isRun){
        if (!stopExecution()){
            return false;
        }
    }
    if (th != NULL){
        free(th);
        th = NULL;
    }
    th = (pthread_t*)malloc(sizeof(pthread_t));
    int state = 0;
    state = pthread_create(th, NULL, run_th, NULL);
    fprintf(stderr, "Thread started\n");
    if (state != 0){
        free(th);
        th = NULL;
        return false;
    }
    struct timespec ts;
    state = clock_gettime(CLOCK_REALTIME, &ts);
    if (state == -1){
        fprintf(stderr, "clock_gettime error\n");
    }
    long ns_max = 1000000000L;
    ts.tv_nsec += (long) (ms % 1000L)*1000000L;
    ts.tv_sec += ts.tv_nsec / ns_max;
    ts.tv_nsec %= ns_max;
    ts.tv_sec += (time_t) ms / 1000;
    
    
    void *res;
    state = pthread_timedjoin_np(*th, &res, &ts);
    if (state != 0){
        fprintf(stderr, "Bad function, try to cancel\n");
        pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
        state = pthread_cancel(*th);
        if (state != 0){
            fprintf(stderr, "Cancel thread error\n");
        }
        fprintf(stderr, "Try to join\n");
        state = pthread_tryjoin_np(*th, &res);
        if (state != 0){
            fprintf(stderr, "Function not joined!\n");
            state = pthread_kill(*th, SIGTERM);
            fprintf(stderr, "Function killed...\n");
        }
        if (res == PTHREAD_CANCELED){
            fprintf(stderr, "thread was canceled\n");
        }else{
            fprintf(stderr, "thread wasn't canceled\n");
        }
    }else{
        fprintf(stderr, "Thread finished\n");
    }
    return true;
}
bool isLoaded(){
    return true;
}

char *getValue(){
    return sg;
}

/*
char * f1(char *s){
    void *lh;
    lh = dlopen(s, RTLD_LAZY);
    if (!lh){
        fprintf(stderr, "dlopen() error: %s\n", dlerror());
        return sg;
    }
    int (*f2)(void);
    f2 = (int (*)(void))dlsym(lh, "f2");
    if (!f2){
        fprintf(stderr, "dlsym() error: %s\n", dlerror());
        return getString();
    }
    f2();
    dlclose(lh);
    return sg;
}*/
#ifdef __cplusplus
}
#endif
